from py4jdbc import settings
from py4jdbc.dbapi2 import connect, Connection
from py4jdbc.gateway_process import GatewayProcess
from py4jdbc.exceptions.dbapi2 import *
from py4jdbc.version import __version__



from collections import ChainMap

from py4jdbc.exceptions import dbapi2
from py4jdbc.exceptions import sqlstate


exc_classes = ChainMap(sqlstate.exc_classes).new_child()


class _PgCodeAgg(dbapi2.CodeAggregatorMeta):
    exc_classes = exc_classes


class ConfigurationFileError(dbapi2.DatabaseError, metaclass=_PgCodeAgg):
    code = 'F0'

class ForeignDataWrapperError(dbapi2.DatabaseError, metaclass=_PgCodeAgg):
    code = 'HV'

class PlOrPgsqlError(dbapi2.DatabaseError, metaclass=_PgCodeAgg):
    code = 'P0'

class InternalError(dbapi2.DatabaseError, metaclass=_PgCodeAgg):
    code = 'XX'


messages = (
    ('00000', 'successful_completion'),
    ('01000', 'warning'),
    ('0100C', 'dynamic_result_sets_returned'),
    ('01008', 'implicit_zero_bit_padding'),
    ('01003', 'null_value_eliminated_in_set_function'),
    ('01007', 'privilege_not_granted'),
    ('01006', 'privilege_not_revoked'),
    ('01004', 'string_data_right_truncation'),
    ('01P01', 'deprecated_feature'),
    ('02000', 'no_data'),
    ('02001', 'no_additional_dynamic_result_sets_returned'),
    ('03000', 'sql_statement_not_yet_complete'),
    ('08000', 'connection_exception'),
    ('08003', 'connection_does_not_exist'),
    ('08006', 'connection_failure'),
    ('08001', 'sqlclient_unable_to_establish_sqlconnection'),
    ('08004', 'sqlserver_rejected_establishment_of_sqlconnection'),
    ('08007', 'transaction_resolution_unknown'),
    ('08P01', 'protocol_violation'),
    ('09000', 'triggered_action_exception'),
    ('0A000', 'feature_not_supported'),
    ('0B000', 'invalid_transaction_initiation'),
    ('0F000', 'locator_exception'),
    ('0F001', 'invalid_locator_specification'),
    ('0L000', 'invalid_grantor'),
    ('0LP01', 'invalid_grant_operation'),
    ('0P000', 'invalid_role_specification'),
    ('0Z000', 'diagnostics_exception'),
    ('0Z002', 'stacked_diagnostics_accessed_without_active_handler'),
    ('20000', 'case_not_found'),
    ('21000', 'cardinality_violation'),
    ('22000', 'data_exception'),
    ('2202E', 'array_subscript_error'),
    ('22021', 'character_not_in_repertoire'),
    ('22008', 'datetime_field_overflow'),
    ('22012', 'division_by_zero'),
    ('22005', 'error_in_assignment'),
    ('2200B', 'escape_character_conflict'),
    ('22022', 'indicator_overflow'),
    ('22015', 'interval_field_overflow'),
    ('2201E', 'invalid_argument_for_logarithm'),
    ('22014', 'invalid_argument_for_ntile_function'),
    ('22016', 'invalid_argument_for_nth_value_function'),
    ('2201F', 'invalid_argument_for_power_function'),
    ('2201G', 'invalid_argument_for_width_bucket_function'),
    ('22018', 'invalid_character_value_for_cast'),
    ('22007', 'invalid_datetime_format'),
    ('22019', 'invalid_escape_character'),
    ('2200D', 'invalid_escape_octet'),
    ('22025', 'invalid_escape_sequence'),
    ('22P06', 'nonstandard_use_of_escape_character'),
    ('22010', 'invalid_indicator_parameter_value'),
    ('22023', 'invalid_parameter_value'),
    ('2201B', 'invalid_regular_expression'),
    ('2201W', 'invalid_row_count_in_limit_clause'),
    ('2201X', 'invalid_row_count_in_result_offset_clause'),
    ('22009', 'invalid_time_zone_displacement_value'),
    ('2200C', 'invalid_use_of_escape_character'),
    ('2200G', 'most_specific_type_mismatch'),
    ('22004', 'null_value_not_allowed'),
    ('22002', 'null_value_no_indicator_parameter'),
    ('22003', 'numeric_value_out_of_range'),
    ('22026', 'string_data_length_mismatch'),
    ('22001', 'string_data_right_truncation'),
    ('22011', 'substring_error'),
    ('22027', 'trim_error'),
    ('22024', 'unterminated_c_string'),
    ('2200F', 'zero_length_character_string'),
    ('22P01', 'floating_point_exception'),
    ('22P02', 'invalid_text_representation'),
    ('22P03', 'invalid_binary_representation'),
    ('22P04', 'bad_copy_file_format'),
    ('22P05', 'untranslatable_character'),
    ('2200L', 'not_an_xml_document'),
    ('2200M', 'invalid_xml_document'),
    ('2200N', 'invalid_xml_content'),
    ('2200S', 'invalid_xml_comment'),
    ('2200T', 'invalid_xml_processing_instruction'),
    ('23000', 'integrity_constraint_violation'),
    ('23001', 'restrict_violation'),
    ('23502', 'not_null_violation'),
    ('23503', 'foreign_key_violation'),
    ('23505', 'unique_violation'),
    ('23514', 'check_violation'),
    ('23P01', 'exclusion_violation'),
    ('24000', 'invalid_cursor_state'),
    ('25000', 'invalid_transaction_state'),
    ('25001', 'active_sql_transaction'),
    ('25002', 'branch_transaction_already_active'),
    ('25008', 'held_cursor_requires_same_isolation_level'),
    ('25003', 'inappropriate_access_mode_for_branch_transaction'),
    ('25004', 'inappropriate_isolation_level_for_branch_transaction'),
    ('25005', 'no_active_sql_transaction_for_branch_transaction'),
    ('25006', 'read_only_sql_transaction'),
    ('25007', 'schema_and_data_statement_mixing_not_supported'),
    ('25P01', 'no_active_sql_transaction'),
    ('25P02', 'in_failed_sql_transaction'),
    ('26000', 'invalid_sql_statement_name'),
    ('27000', 'triggered_data_change_violation'),
    ('28000', 'invalid_authorization_specification'),
    ('28P01', 'invalid_password'),
    ('2B000', 'dependent_privilege_descriptors_still_exist'),
    ('2BP01', 'dependent_objects_still_exist'),
    ('2D000', 'invalid_transaction_termination'),
    ('2F000', 'sql_routine_exception'),
    ('2F005', 'function_executed_no_return_statement'),
    ('2F002', 'modifying_sql_data_not_permitted'),
    ('2F003', 'prohibited_sql_statement_attempted'),
    ('2F004', 'reading_sql_data_not_permitted'),
    ('34000', 'invalid_cursor_name'),
    ('38000', 'external_routine_exception'),
    ('38001', 'containing_sql_not_permitted'),
    ('38002', 'modifying_sql_data_not_permitted'),
    ('38003', 'prohibited_sql_statement_attempted'),
    ('38004', 'reading_sql_data_not_permitted'),
    ('39000', 'external_routine_invocation_exception'),
    ('39001', 'invalid_sqlstate_returned'),
    ('39004', 'null_value_not_allowed'),
    ('39P01', 'trigger_protocol_violated'),
    ('39P02', 'srf_protocol_violated'),
    ('3B000', 'savepoint_exception'),
    ('3B001', 'invalid_savepoint_specification'),
    ('3D000', 'invalid_catalog_name'),
    ('3F000', 'invalid_schema_name'),
    ('40000', 'transaction_rollback'),
    ('40002', 'transaction_integrity_constraint_violation'),
    ('40001', 'serialization_failure'),
    ('40003', 'statement_completion_unknown'),
    ('40P01', 'deadlock_detected'),
    ('42000', 'syntax_error_or_access_rule_violation'),
    ('42601', 'syntax_error'),
    ('42501', 'insufficient_privilege'),
    ('42846', 'cannot_coerce'),
    ('42803', 'grouping_error'),
    ('42P20', 'windowing_error'),
    ('42P19', 'invalid_recursion'),
    ('42830', 'invalid_foreign_key'),
    ('42602', 'invalid_name'),
    ('42622', 'name_too_long'),
    ('42939', 'reserved_name'),
    ('42804', 'datatype_mismatch'),
    ('42P18', 'indeterminate_datatype'),
    ('42P21', 'collation_mismatch'),
    ('42P22', 'indeterminate_collation'),
    ('42809', 'wrong_object_type'),
    ('42703', 'undefined_column'),
    ('42883', 'undefined_function'),
    ('42P01', 'undefined_table'),
    ('42P02', 'undefined_parameter'),
    ('42704', 'undefined_object'),
    ('42701', 'duplicate_column'),
    ('42P03', 'duplicate_cursor'),
    ('42P04', 'duplicate_database'),
    ('42723', 'duplicate_function'),
    ('42P05', 'duplicate_prepared_statement'),
    ('42P06', 'duplicate_schema'),
    ('42P07', 'duplicate_table'),
    ('42712', 'duplicate_alias'),
    ('42710', 'duplicate_object'),
    ('42702', 'ambiguous_column'),
    ('42725', 'ambiguous_function'),
    ('42P08', 'ambiguous_parameter'),
    ('42P09', 'ambiguous_alias'),
    ('42P10', 'invalid_column_reference'),
    ('42611', 'invalid_column_definition'),
    ('42P11', 'invalid_cursor_definition'),
    ('42P12', 'invalid_database_definition'),
    ('42P13', 'invalid_function_definition'),
    ('42P14', 'invalid_prepared_statement_definition'),
    ('42P15', 'invalid_schema_definition'),
    ('42P16', 'invalid_table_definition'),
    ('42P17', 'invalid_object_definition'),
    ('44000', 'with_check_option_violation'),
    ('53000', 'insufficient_resources'),
    ('53100', 'disk_full'),
    ('53200', 'out_of_memory'),
    ('53300', 'too_many_connections'),
    ('53400', 'configuration_limit_exceeded'),
    ('54000', 'program_limit_exceeded'),
    ('54001', 'statement_too_complex'),
    ('54011', 'too_many_columns'),
    ('54023', 'too_many_arguments'),
    ('55000', 'object_not_in_prerequisite_state'),
    ('55006', 'object_in_use'),
    ('55P02', 'cant_change_runtime_param'),
    ('55P03', 'lock_not_available'),
    ('57000', 'operator_intervention'),
    ('57014', 'query_canceled'),
    ('57P01', 'admin_shutdown'),
    ('57P02', 'crash_shutdown'),
    ('57P03', 'cannot_connect_now'),
    ('57P04', 'database_dropped'),
    ('58000', 'system_error'),
    ('58030', 'io_error'),
    ('58P01', 'undefined_file'),
    ('58P02', 'duplicate_file'),
    ('F0000', 'config_file_error'),
    ('F0001', 'lock_file_exists'),
    ('HV000', 'fdw_error'),
    ('HV005', 'fdw_column_name_not_found'),
    ('HV002', 'fdw_dynamic_parameter_value_needed'),
    ('HV010', 'fdw_function_sequence_error'),
    ('HV021', 'fdw_inconsistent_descriptor_information'),
    ('HV024', 'fdw_invalid_attribute_value'),
    ('HV007', 'fdw_invalid_column_name'),
    ('HV008', 'fdw_invalid_column_number'),
    ('HV004', 'fdw_invalid_data_type'),
    ('HV006', 'fdw_invalid_data_type_descriptors'),
    ('HV091', 'fdw_invalid_descriptor_field_identifier'),
    ('HV00B', 'fdw_invalid_handle'),
    ('HV00C', 'fdw_invalid_option_index'),
    ('HV00D', 'fdw_invalid_option_name'),
    ('HV090', 'fdw_invalid_string_length_or_buffer_length'),
    ('HV00A', 'fdw_invalid_string_format'),
    ('HV009', 'fdw_invalid_use_of_null_pointer'),
    ('HV014', 'fdw_too_many_handles'),
    ('HV001', 'fdw_out_of_memory'),
    ('HV00P', 'fdw_no_schemas'),
    ('HV00J', 'fdw_option_name_not_found'),
    ('HV00K', 'fdw_reply_handle'),
    ('HV00Q', 'fdw_schema_not_found'),
    ('HV00R', 'fdw_table_not_found'),
    ('HV00L', 'fdw_unable_to_create_execution'),
    ('HV00M', 'fdw_unable_to_create_reply'),
    ('HV00N', 'fdw_unable_to_establish_connection'),
    ('P0000', 'plpgsql_error'),
    ('P0001', 'raise_exception'),
    ('P0002', 'no_data_found'),
    ('P0003', 'too_many_rows'),
    ('XX000', 'internal_error'),
    ('XX001', 'data_corrupted'),
    ('XX002', 'index_corrupted'),
)


import re
import logging

from functools import partial

from py4jdbc.utils import CachedAttr
from py4jdbc.resultset import ResultSet, ColumnResultSet
from py4jdbc.gateway_process import GatewayProcess
from py4jdbc.exceptions import dbapi2 as dbapi2_exc
from py4jdbc.exceptions.py4j import reraise_jvm_exception


paramstyle = 'qmark'

# DB-API 2.0 Module Interface connect constructor
def connect(jdbc_url, user=None, password=None, gateway=None, autocommit=None):
    return Connection(jdbc_url, user, password, gateway, autocommit=None)


class _ExceptionMixin:
    '''Exposes dbapi2 exceptions classes as class-level attributes.
    '''
    Error = dbapi2_exc.Error
    Warning = dbapi2_exc.Warning
    InterfaceError = dbapi2_exc.InterfaceError
    DatabaseError = dbapi2_exc.DatabaseError
    InternalError = dbapi2_exc.InternalError
    OperationalError = dbapi2_exc.OperationalError
    ProgrammingError = dbapi2_exc.ProgrammingError
    IntegrityError = dbapi2_exc.IntegrityError
    DataError = dbapi2_exc.DataError
    NotSupportedError = dbapi2_exc.NotSupportedError


class _JdbcPropertyAlias:

    def __init__(self, property_name):
        self.name = property_name

    def __get__(self, inst, Cls):
        return inst.get_property(self.name)

    def __set__(self, inst, value):
        inst.set_property(self.name, value)


# DB-API 2.0 Connection Object
class Connection(_ExceptionMixin):

    autocommit = _JdbcPropertyAlias('AutoCommit')

    def __init__(self, jdbc_url, user=None, password=None, gateway=None, autocommit=None):
        self._gateway_arg = gateway
        self._jdbc_url = jdbc_url
        self._user = user
        self._password = password
        self._closed = False
        self._logger = logging.getLogger('py4jdbc')
        self._autostarted_gateway = False

        # Set autocommit
        if autocommit is not None:
            self.autocommit = autocommit

    def __enter__(self):
        if not self._gateway.is_running:
            self._gateway.run()
            self._autostarted_gateway = True
        return self

    def __exit__(self, *args):
        if not self._closed:
            self.close()
        # If this connection started a gateway server, shut it down now.
        if self._gateway.is_running and self._autostarted_gateway:
            msg = "Shutting down GatewayProcess auto-started by connection %r"
            self._logger.info(msg, self)
            self._gateway.shutdown()
        else:
            msg = ("GatewayProcess wasn't auto-started by connection. "
                   "Leaving it up.")
            self._logger.info(msg)

    # -----------------------------------------------------------------------
    # Py4j accessors.
    # -----------------------------------------------------------------------
    @property
    def _entry_point(self):
        return self._gateway.entry_point

    _ep = _entry_point

    @CachedAttr
    def _gateway_process(self):
        proc = GatewayProcess(require_ctx_manager=False)
        return proc

    @CachedAttr
    def _gateway(self):
        gw = self._gateway_arg
        if gw is not None:
            # If a gateway process was passed in, get the actual gateway.
            if isinstance(gw, GatewayProcess):
                gw = self._gateway_arg.gateway
            self._gateway = gw
            return gw
        return self._gateway_process.gateway

    # -----------------------------------------------------------------------
    # Java connection accessors.
    # -----------------------------------------------------------------------
    def get_property(self, name: str):
        method_name = 'get' + name
        if method_name in dir(self._jconn):
            method = getattr(self._jconn, method_name)
        else:
            method = partial(self._jconn.getProperty, name)
        with reraise_jvm_exception(self._gateway):
            return method()

    def set_property(self, name: str, value: str):
        method_name = 'set' + name
        if method_name in dir(self._jconn):
            method = getattr(self._jconn, method_name)
        else:
            method = partial(self._jconn.setProperty, name)
        with reraise_jvm_exception(self._gateway):
            return method(value)

    @CachedAttr
    def _py4jdbc_connection(self):
        self._logger.debug('Connecting as: %r', self._user, self._jdbc_url)
        return self._ep.getConnection(self._jdbc_url, self._user, self._password)

    @property
    def _conn(self):
        return self._py4jdbc_connection

    @CachedAttr
    def _jdbc_connection(self):
        return self._py4jdbc_connection.getJdbcConnection()

    @property
    def _jconn(self):
        return self._jdbc_connection

    def close(self):
        if self._closed:
            raise self.Error("Connection is already closed.")
        with reraise_jvm_exception(self._gateway):
            self._jconn.close()
            self._closed = True

    def commit(self):
        with reraise_jvm_exception(self._gateway):
            self._jconn.commit()

    def rollback(self):
        with reraise_jvm_exception(self._gateway):
            self._jconn.rollback()

    def cursor(self):
        return Cursor(self)

    # ------------------------------------------------------------------------
    # Introspection
    # ------------------------------------------------------------------------
    @CachedAttr
    def _metadata(self):
        return self._jdbc_connection.getMetaData()

    def get_tables(self, catalog=None, schema=None, table='%'):
        rs = self._metadata.getTables(catalog, schema, table, None)
        rs = self._ep.mkPyResultSet(rs)
        return ResultSet(rs, gateway=self._gateway, case_insensitive=True)

    def get_columns(self, catalog=None, schema=None, table='%', column=None):
        rs = self._metadata.getColumns(catalog, schema, table, column)
        rs = self._ep.mkPyResultSet(rs)
        return ColumnResultSet(rs, gateway=self._gateway)

    def get_primary_keys(self, catalog=None, schema=None, table='%', column=None):
        rs = self._metadata.getPrimaryKeys(catalog, schema, table)
        rs = self._ep.mkPyResultSet(rs)
        return ColumnResultSet(rs, gateway=self._gateway)

    def get_imported_keys(self, catalog=None, schema=None, table='%', column=None):
        rs = self._metadata.getImportedKeys(catalog, schema, table)
        rs = self._ep.mkPyResultSet(rs)
        return ColumnResultSet(rs, gateway=self._gateway)

    def get_exported_keys(self, catalog=None, schema=None, table='%', column=None):
        rs = self._metadata.getExportedKeys(catalog, schema, table)
        rs = self._ep.mkPyResultSet(rs)
        return ColumnResultSet(rs, gateway=self._gateway)

    def get_privileges(self, catalog=None, schema=None, table='%', column=None):
        rs = self._metadata.getTablePrivileges(catalog, schema, table)
        rs = self._ep.mkPyResultSet(rs)
        return ColumnResultSet(rs, gateway=self._gateway)


# DB-API 2.0 Cursor Object
class Cursor(_ExceptionMixin):

    def __init__(self, connection):
        self.connection = connection
        self._reset()
        self._logger = logging.getLogger('py4jdbc')

    def _reset(self):
        self._closed = False
        self._rs = None
        self._description = None

    # ------------------------------------------------------------------------
    # Shortcuts for reaching java objects.
    # ------------------------------------------------------------------------
    @property
    def _gateway(self):
        return self.connection._gateway

    @property
    def _entrypoint(self):
        return self.connection._entry_point

    _ep = _entrypoint

    @property
    def _jdbc_connection(self):
        return self.connection._jdbc_connection

    _jconn = _jdbc_connection

    # ------------------------------------------------------------------------
    # Actual dbapi 2.0 methods.
    # ------------------------------------------------------------------------
    @property
    def description(self):
        if self._description is None:
            self._description = self._rs.description()
        return self._description

    def close(self):
        if self._closed:
            raise self.Error("Connection is already closed.")
        self._logger.info('Closing cursor')
        self._close_last()
        self._closed = True

    def _close_last(self):
        if self._rs is not None:
            with reraise_jvm_exception(self._gateway):
                self._rs.close()

    def execute(self, operation, parameters=None):
        if self._closed:
            raise self.Error("Connection is closed.")
        with reraise_jvm_exception(self._gateway):
            if parameters is None:
                rs = self.connection._conn.execute(operation)
            else:
                rs = self.connection._conn.execute(operation, parameters)
            if rs is None:
                del self._rs
                return
            self._rs = ResultSet(rs, gateway=self._gateway)
            return self._rs

    def executemany(self, operation, parameter_seq):
        if self._closed:
            raise self.Error("Connection is closed.")
        with reraise_jvm_exception(self._gateway):
            rs = self.connection._conn.executeMany(operation, parameter_seq)
            if rs is None:
                del self._rs
                return
            self._rs = ResultSet(rs, gateway=self._gateway)
            return self._rs

    def fetchone(self):
        if self._closed:
            raise self.Error("Connection is closed.")
        return self._rs.fetchone()

    def fetchmany(self, size=None):
        if self._closed:
            raise self.Error("Connection is closed.")
        return self._rs.fetchmany(size)

    def fetchall(self):
        if self._closed:
            raise self.Error("Connection is closed.")
        return self._rs.fetchall()



import io
import pickle
import logging
import collections

from py4jdbc.utils import CachedAttr
from py4jdbc.java_sql_types import java_sql_types
from py4jdbc.exceptions.py4j import reraise_jvm_exception


class ResultSet:

    def __init__(self, rs, gateway, batchsize=500,
                 fieldnames_attr='column_names', case_insensitive=False):
        self._gateway = gateway
        self._rs = rs
        self._fieldnames_attr = fieldnames_attr
        self._logger = logging.getLogger('py4jdbc')
        self._case_insensitive = case_insensitive
        self.index = 0
        self.batchsize = 500

    def metadata(self):
        '''Get all resultset metadata as a dictionary.
        '''
        meta = self._rs.getMetadataMap()
        return meta

    def close(self):
        self._logger.debug('Closing %r', self)
        self._rs.close()

    def _make_rows(self, bytes):
        values = pickle.loads(bytes)
        rows = tuple(map(self.Row._make, values))
        self.index += len(rows)
        return rows

    @CachedAttr
    def rows(self):
        return self.fetchall()

    @CachedAttr
    def column_names(self):
        self._logger.info('Fetching column names: %r', self)
        with reraise_jvm_exception(self._gateway):
            resp = self._rs.colNames()
        return pickle.loads(resp)

    @property
    def column_labels(self):
        self._logger.debug('Fetching column labels: %r', self)
        with reraise_jvm_exception(self._gateway):
            resp = self._rs.colLabels()
        return pickle.loads(resp)

    @CachedAttr
    def Row(self):
       field_names = getattr(self, self._fieldnames_attr)
       return make_row(field_names, case_insensitive=self._case_insensitive)

    def fetchone(self):
        self._logger.debug('Fetching one: %r', self)
        vals = pickle.loads(self._rs.fetchOne())
        if vals:
            return self.Row(*vals)

    def fetchmany(self, size=None):
        self._logger.debug("Batchsize is %d", size)
        with reraise_jvm_exception(self._gateway):
            resp = self._rs.fetchMany(size)
        return self._make_rows(resp)

    def fetchall(self):
        msg = ('Calling fetching all can cause memory errors '
               'in the py4jdbc gateway server. Consider using fetchmany '
               'instead.')
        self._logger.warning(msg)
        with reraise_jvm_exception(self._gateway):
            resp = self._rs.fetchAll()
        return self._make_rows(resp)

    def description(self):
        self._logger.debug('Getting cursor description.')
        with reraise_jvm_exception(self._gateway):
            resp = self._rs.getDescription()
        return pickle.loads(resp)

    def next(self):
        return self.fetchone()

    def __iter__(self):
        self._logger.debug('Iterating over result set: %r', self)
        while True:
            batch = self.fetchmany(self.batchsize)
            if not batch:
                self._logger.debug('Done iterating over result set: %r', self)
                return
            for row in batch:
                yield row


class ColumnResultSet(ResultSet):

    _type_map = dict(item[::-1] for item in java_sql_types)

    @CachedAttr
    def Row(self):
        field_names = getattr(self, self._fieldnames_attr)
        BaseRow = make_row(field_names, case_insensitive=True)
        class ColumnRow(BaseRow):
            _type_map = self._type_map
            @property
            def data_type(self):
                return self._type_map[self.DATA_TYPE]
        return ColumnRow

# ------------------------------------------------------------------------------
# Row base class.
# ------------------------------------------------------------------------------
class _RowFieldAccessor:
    '''Used to lokup values in the row tuple.
    '''
    def __init__(self, fieldname):
        self.fieldname = fieldname

    def __get__(self, inst, Cls=None):
        idx = inst._fieldnames.index(self.fieldname)
        return inst[idx]

    __getitem__ = object.__getattribute__


class _Row(tuple):
    '''A modified namedtuple supports dict-style value access and doesn't
    mangle field names or complain about fieldnames that clash with Python
    keywords. Subclasses must supply _fieldnames and __slots__
    '''
    _fieldnames = None
    __slots__ = tuple()

    def __new__(Cls, *values):
        return tuple.__new__(Cls, values)

    @classmethod
    def _make(Cls, iterable):
        return tuple.__new__(Cls, iterable)

    def _replace(self, **kwds):
        'Return a new a object replacing specified fields with new values'
        result = self._make(map(kwds.pop, self._fieldnames, self))
        if kwds:
            raise ValueError('Got unexpected field names: %r' % list(kwds))
        return result

    def __repr__(self):
        'Return a nicely formatted representation string'
        return '%s(%r)' % (self.__class__.__name__, self._asdict())

    def _asdict(self):
        'Return a new OrderedDict which maps field names to their values.'
        return collections.OrderedDict(zip(self._fieldnames, self))

    def __getnewargs__(self):
        'Return self as a plain tuple.  Used by copy and pickle.'
        return tuple(self)


def make_row(fieldnames, case_insensitive=False):
    '''Return a customized namedtuple variant that supports lower-case
    access of all values by column name.
    '''
    flds = tuple(fieldnames)
    lc_flds = tuple(s.lower() for s in flds)
    slots = _Row.__slots__ + flds + lc_flds

    members = [('_fieldnames', flds)]
    for field in flds:
        members.append((field, _RowFieldAccessor(field)))
        # This effectively makes lookups case-insensitive.
        if case_insensitive:
            members.append((field.lower(), _RowFieldAccessor(field)))

    Row = type('Row', (_Row,), dict(members))

    return Row


# class ResultSetMeta:

#     def __init__(self, meta):
#         self._meta = meta

#     @property
#     def column_no_nulls(self):
#         '''The constant indicating that a column does not allow NULL value
#         '''
#         return self._meta.columnNoNulls()

#     @property
#     def column_nullable(self):
#         '''The constant indicating that a column allows NULL value
#         '''
#         return self._meta.columnNullable()

#     @property
#     def column_nullable_unknown(self):
#         ''' indicating that the nullability of a column's values is unknown.
#         '''
#         return self._meta.columnNullableUnknown()

#     def get_catalog_name(self, column):
#         '''Gets the designated column's table's catalog name.
#         '''
#         return self._meta.getCatalogName(column)

#     def get_column_class_name(self, column):
#         '''Returns the fully-qualified name of the Java class whose instances
#         are manufactured if the method ResultSet.getObject is called to
#         retrieve a value from the column.'''
#         return self._meta.getColumnClassName(column)

#     def get_column_count(self):
#         '''Returns the number of columns in this ResultSet object.
#         '''
#         return self._meta.getColumnCount()

#     def get_column_display_size(self, column):
#         '''Indicates the designated column's normal maximum width in
#         characters.
#         '''
#         return self._meta.getColumnDisplaySize(column)

#     def get_column_label(self, column):
#         '''Gets the designated column's suggested title for use in printouts
#         and displays.
#         '''
#         return self._meta.getColumnLabel(column)

#     def get_column_name(self, column):
#         '''Get the designated column's name.
#         '''

#         return self._meta.getColumnName(column)

#     def get_column_type(self, column):
#         '''Retrieves the designated column's SQL type.
#         '''
#         return self._meta.getColumnType(column)

#     def get_column_type_name(self, column):
#         '''Retrieves the designated column's database-specific type name.
#         '''
#         return self._meta.getColumnTypeName(column)

#     def get_precision(self, column):
#         '''Get the designated column's specified column size.
#         '''
#         return self._meta.getPrecision(column)

#     def get_scale(self, column):
#         '''Gets the designated column's number of digits to right of the
#         decimal point.
#         '''
#         return self._meta.getScale(column)

#     def get_schema_name(self, column):
#         '''Get the designated column's table's schema.
#         '''
#         return self._meta.getSchemaName(column)

#     def get_table_name(self, column):
#         '''Gets the designated column's table name.
#         '''
#         return self._meta.getTableName(column)

#     def is_auto_increment(self, column):
#         '''Indicates whether the designated column is automatically numbered.
#         '''
#         return self._meta.isAutoIncrement(column)

#     def is_case_sensitive(self, column):
#         '''Indicates whether a column's case matters.
#         '''

#         return self._meta.isCaseSensitive(column)

#     def is_currency(self, column):
#         '''Indicates whether the designated column is a cash value.
#         '''
#         return self._meta.isCurrency(column)

#     def is_definitely_writable(self, column):
#         '''Indicates whether a write on the designated column will definitely
#         succeed.
#         '''
#         return self._meta.isDefinitelyWritable(column)

#     def is_nullable(self, column):
#         '''Indicates the nullability of values in the designated column.
#         '''
#         return self._meta.isNullable(column)

#     def is_read_only(self, column):(column)
#         '''Indicates whether the designated column is definitely not writable.
#         '''
#         return self._meta.isReadOnly(column)

#     def is_searchable(self, column):
#         '''Indicates whether the designated column can be used in a where
#         clause.
#         '''
#         return self._meta.isSearchable(column)

#     def is_signed(self, column):
#         '''Indicates whether values in the designated column are signed
#         numbers.
#         '''
#         return self._meta.isSigned(column)

#     def is_writable(self, column):
#         '''Indicates whether it is possible for a write on the designated
#         column to succeed.
#         '''
#         return self._meta.isWritable(column)




java_sql_types = [
    ('ARRAY', 2003),
    ('BIGINT', -5),
    ('BINARY', -2),
    ('BIT', -7),
    ('BLOB', 2004),
    ('BOOLEAN', 16),
    ('CHAR', 1),
    ('CLOB', 2005),
    ('DATALINK', 70),
    ('DATE', 91),
    ('DECIMAL', 3),
    ('DISTINCT', 2001),
    ('DOUBLE', 8),
    ('FLOAT', 6),
    ('INTEGER', 4),
    ('JAVA_OBJECT', 2000),
    ('LONGNVARCHAR', -16),
    ('LONGVARBINARY', -4),
    ('LONGVARCHAR', -1),
    ('NCHAR', -15),
    ('NCLOB', 2011),
    ('NULL', 0),
    ('NUMERIC', 2),
    ('NVARCHAR', -9),
    ('OTHER', 1111),
    ('REAL', 7),
    ('REF', 2006),
    ('ROWID', -8),
    ('SMALLINT', 5),
    ('SQLXML', 2009),
    ('STRUCT', 2002),
    ('TIME', 92),
    ('TIMESTAMP', 93),
    ('TINYINT', -6),
    ('VARBINARY', -3),
    ('VARCHAR', 12),
]

from enum import Enum


class Connection(Enum):
    '''Constants for module java.sql.Connection.
    '''
    module = "java.sql.Connection"

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Connection.html#TRANSACTION_NONE
    # modifiers: public static final int
    TRANSACTION_NONE = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Connection.html#TRANSACTION_READ_COMMITTED
    # modifiers: public static final int
    TRANSACTION_READ_COMMITTED = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Connection.html#TRANSACTION_READ_UNCOMMITTED
    # modifiers: public static final int
    TRANSACTION_READ_UNCOMMITTED = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Connection.html#TRANSACTION_REPEATABLE_READ
    # modifiers: public static final int
    TRANSACTION_REPEATABLE_READ = 4

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Connection.html#TRANSACTION_SERIALIZABLE
    # modifiers: public static final int
    TRANSACTION_SERIALIZABLE = 8


class Databasemetadata(Enum):
    '''Constants for module java.sql.DatabaseMetaData.
    '''
    module = "java.sql.DatabaseMetaData"

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#attributeNoNulls
    # modifiers: public static final short
    attributeNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#attributeNullable
    # modifiers: public static final short
    attributeNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#attributeNullableUnknown
    # modifiers: public static final short
    attributeNullableUnknown = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#bestRowNotPseudo
    # modifiers: public static final int
    bestRowNotPseudo = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#bestRowPseudo
    # modifiers: public static final int
    bestRowPseudo = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#bestRowSession
    # modifiers: public static final int
    bestRowSession = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#bestRowTemporary
    # modifiers: public static final int
    bestRowTemporary = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#bestRowTransaction
    # modifiers: public static final int
    bestRowTransaction = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#bestRowUnknown
    # modifiers: public static final int
    bestRowUnknown = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#columnNoNulls
    # modifiers: public static final int
    columnNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#columnNullable
    # modifiers: public static final int
    columnNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#columnNullableUnknown
    # modifiers: public static final int
    columnNullableUnknown = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionColumnIn
    # modifiers: public static final int
    functionColumnIn = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionColumnInOut
    # modifiers: public static final int
    functionColumnInOut = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionColumnOut
    # modifiers: public static final int
    functionColumnOut = 3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionColumnResult
    # modifiers: public static final int
    functionColumnResult = 5

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionColumnUnknown
    # modifiers: public static final int
    functionColumnUnknown = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionNoNulls
    # modifiers: public static final int
    functionNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionNoTable
    # modifiers: public static final int
    functionNoTable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionNullable
    # modifiers: public static final int
    functionNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionNullableUnknown
    # modifiers: public static final int
    functionNullableUnknown = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionResultUnknown
    # modifiers: public static final int
    functionResultUnknown = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionReturn
    # modifiers: public static final int
    functionReturn = 4

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#functionReturnsTable
    # modifiers: public static final int
    functionReturnsTable = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeyCascade
    # modifiers: public static final int
    importedKeyCascade = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeyInitiallyDeferred
    # modifiers: public static final int
    importedKeyInitiallyDeferred = 5

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeyInitiallyImmediate
    # modifiers: public static final int
    importedKeyInitiallyImmediate = 6

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeyNoAction
    # modifiers: public static final int
    importedKeyNoAction = 3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeyNotDeferrable
    # modifiers: public static final int
    importedKeyNotDeferrable = 7

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeyRestrict
    # modifiers: public static final int
    importedKeyRestrict = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeySetDefault
    # modifiers: public static final int
    importedKeySetDefault = 4

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#importedKeySetNull
    # modifiers: public static final int
    importedKeySetNull = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureColumnIn
    # modifiers: public static final int
    procedureColumnIn = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureColumnInOut
    # modifiers: public static final int
    procedureColumnInOut = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureColumnOut
    # modifiers: public static final int
    procedureColumnOut = 4

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureColumnResult
    # modifiers: public static final int
    procedureColumnResult = 3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureColumnReturn
    # modifiers: public static final int
    procedureColumnReturn = 5

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureColumnUnknown
    # modifiers: public static final int
    procedureColumnUnknown = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureNoNulls
    # modifiers: public static final int
    procedureNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureNoResult
    # modifiers: public static final int
    procedureNoResult = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureNullable
    # modifiers: public static final int
    procedureNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureNullableUnknown
    # modifiers: public static final int
    procedureNullableUnknown = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureResultUnknown
    # modifiers: public static final int
    procedureResultUnknown = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#procedureReturnsResult
    # modifiers: public static final int
    procedureReturnsResult = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#sqlStateSQL
    # modifiers: public static final int
    sqlStateSQL = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#sqlStateSQL99
    # modifiers: public static final int
    sqlStateSQL99 = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#sqlStateXOpen
    # modifiers: public static final int
    sqlStateXOpen = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#tableIndexClustered
    # modifiers: public static final short
    tableIndexClustered = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#tableIndexHashed
    # modifiers: public static final short
    tableIndexHashed = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#tableIndexOther
    # modifiers: public static final short
    tableIndexOther = 3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#tableIndexStatistic
    # modifiers: public static final short
    tableIndexStatistic = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typeNoNulls
    # modifiers: public static final int
    typeNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typeNullable
    # modifiers: public static final int
    typeNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typeNullableUnknown
    # modifiers: public static final int
    typeNullableUnknown = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typePredBasic
    # modifiers: public static final int
    typePredBasic = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typePredChar
    # modifiers: public static final int
    typePredChar = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typePredNone
    # modifiers: public static final int
    typePredNone = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#typeSearchable
    # modifiers: public static final int
    typeSearchable = 3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#versionColumnNotPseudo
    # modifiers: public static final int
    versionColumnNotPseudo = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#versionColumnPseudo
    # modifiers: public static final int
    versionColumnPseudo = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html#versionColumnUnknown
    # modifiers: public static final int
    versionColumnUnknown = 0


class Parametermetadata(Enum):
    '''Constants for module java.sql.ParameterMetaData.
    '''
    module = "java.sql.ParameterMetaData"

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterModeIn
    # modifiers: public static final int
    parameterModeIn = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterModeInOut
    # modifiers: public static final int
    parameterModeInOut = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterModeOut
    # modifiers: public static final int
    parameterModeOut = 4

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterModeUnknown
    # modifiers: public static final int
    parameterModeUnknown = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterNoNulls
    # modifiers: public static final int
    parameterNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterNullable
    # modifiers: public static final int
    parameterNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ParameterMetaData.html#parameterNullableUnknown
    # modifiers: public static final int
    parameterNullableUnknown = 2


class Resultset(Enum):
    '''Constants for module java.sql.ResultSet.
    '''
    module = "java.sql.ResultSet"

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#CLOSE_CURSORS_AT_COMMIT
    # modifiers: public static final int
    CLOSE_CURSORS_AT_COMMIT = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#CONCUR_READ_ONLY
    # modifiers: public static final int
    CONCUR_READ_ONLY = 1007

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#CONCUR_UPDATABLE
    # modifiers: public static final int
    CONCUR_UPDATABLE = 1008

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#FETCH_FORWARD
    # modifiers: public static final int
    FETCH_FORWARD = 1000

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#FETCH_REVERSE
    # modifiers: public static final int
    FETCH_REVERSE = 1001

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#FETCH_UNKNOWN
    # modifiers: public static final int
    FETCH_UNKNOWN = 1002

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#HOLD_CURSORS_OVER_COMMIT
    # modifiers: public static final int
    HOLD_CURSORS_OVER_COMMIT = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#TYPE_FORWARD_ONLY
    # modifiers: public static final int
    TYPE_FORWARD_ONLY = 1003

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#TYPE_SCROLL_INSENSITIVE
    # modifiers: public static final int
    TYPE_SCROLL_INSENSITIVE = 1004

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSet.html#TYPE_SCROLL_SENSITIVE
    # modifiers: public static final int
    TYPE_SCROLL_SENSITIVE = 1005


class Resultsetmetadata(Enum):
    '''Constants for module java.sql.ResultSetMetaData.
    '''
    module = "java.sql.ResultSetMetaData"

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSetMetaData.html#columnNoNulls
    # modifiers: public static final int
    columnNoNulls = 0

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSetMetaData.html#columnNullable
    # modifiers: public static final int
    columnNullable = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/ResultSetMetaData.html#columnNullableUnknown
    # modifiers: public static final int
    columnNullableUnknown = 2


class Statement(Enum):
    '''Constants for module java.sql.Statement.
    '''
    module = "java.sql.Statement"

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#CLOSE_ALL_RESULTS
    # modifiers: public static final int
    CLOSE_ALL_RESULTS = 3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#CLOSE_CURRENT_RESULT
    # modifiers: public static final int
    CLOSE_CURRENT_RESULT = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#EXECUTE_FAILED
    # modifiers: public static final int
    EXECUTE_FAILED = -3

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#KEEP_CURRENT_RESULT
    # modifiers: public static final int
    KEEP_CURRENT_RESULT = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#NO_GENERATED_KEYS
    # modifiers: public static final int
    NO_GENERATED_KEYS = 2

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#RETURN_GENERATED_KEYS
    # modifiers: public static final int
    RETURN_GENERATED_KEYS = 1

    # https://docs.oracle.com/javase/7/docs/api/java/sql/Statement.html#SUCCESS_NO_INFO
    # modifiers: public static final int
    SUCCESS_NO_INFO = -2



import pickle
import functools

from py4j.protocol import Py4JJavaError


# DB-API 2.0 Module Interface Exceptions
class Error(Exception):
    pass

class JavaError(Error):

    def __init__(self, java_exc, *args, **kwargs):
        super().__init__(java_exc.tb_stripped(), *args, **kwargs)
        self.java_exc = java_exc

class Warning(JavaError):
    pass

class InterfaceError(JavaError):
    pass

class DatabaseError(JavaError):
    pass

class InternalError(DatabaseError):
    pass

class OperationalError(DatabaseError):
    pass

class ProgrammingError(DatabaseError):
    pass

class IntegrityError(DatabaseError):
    pass

class DataError(DatabaseError):
    pass

class NotSupportedError(DatabaseError):
    pass


class CodeAggregatorMeta(type):
    exc_classes = None

    def __new__(meta, name, bases, nmspc):
         Cls = super().__new__(meta, name, bases, nmspc)
         meta.exc_classes[Cls.code] = Cls
         return Cls




import os
from os.path import abspath, dirname, join
import logging.config


PACKAGE = dirname(abspath(__file__))

LOGLEVEL = os.getenv('PY4JDBC_LOGLEVEL', 'INFO')
USE_COLOR = os.getenv('PY4JDBC_LOG_COLOR', 1) not in ('0', 'False')

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': "%(asctime)s [%(name)s] %(levelname)s %(module)s::%(funcName)s: %(message)s",
            'datefmt': '%H:%M:%S'
        }
    },
    'handlers': {
        'default': {'level': 'INFO', 'formatter': 'standard',
                    'class': 'logging.StreamHandler'},
    },
    'loggers': {
        'py4jdbc': {
            'handlers': ['default'], 'level': LOGLEVEL, 'propagate': True
        },
    },
}

if USE_COLOR:
    LOGGING_CONFIG['handlers']['default']['class'] = 'py4jdbc.ansiterm.ColorizingStreamHandler'

logging.config.dictConfig(LOGGING_CONFIG)



class CachedAttr(object):
    '''Computes attribute value and caches it in instance.

    Example:
        class MyClass(object):
            def myMethod(self):
                # ...
            myMethod = Cached(myMethod)
    Use "del inst.myMethod" to clear cache.
    http://code.activestate.com/recipes/276643/
    '''
    def __init__(self, method, name=None):
        self.method = method
        self.name = name or method.__name__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        result = self.method(inst)
        setattr(inst, self.name, result)
        return result


def resolve_name(name, module_name=None, raise_exc=False):
    '''Given a name string and module prefix, try to import the name.
    '''
    if not isinstance(name, str):
        raise TypeError('Need a stringy name to import.')
    if module_name is None:
        module_name, _, name = name.rpartition('.')
    try:
        module = __import__(module_name, globals(), locals(), [name], 0)
    except ImportError as exc:
        if raise_exc:
            raise exc
    else:
        return getattr(module, name)


import re
import pickle
import functools
import contextlib

import py4j.protocol

from py4jdbc.utils import CachedAttr, resolve_name
from py4jdbc.exceptions.dbapi2 import DatabaseError


_java_error_type_map = {
    'org.postgresql': 'py4jdbc.exceptions.postgres',
    'com.vertica': 'py4jdbc.exceptions.vertica',
    'java.sql': 'py4jdbc.exceptions.sqlstate',
}


@contextlib.contextmanager
def reraise_jvm_exception(gateway_client):
    try:
        yield
    except py4j.protocol.Py4JJavaError as py4j_exc:
        jexc = get_jvm_exc_data(py4j_exc.java_exception, gateway_client)

        # Special handling only applies to java.sql.SQLException
        if not jexc.is_java_sql_error():
            raise py4j_exc

        # Now find the corresponding py4jdbc exception and raise that one
        # from the Py4j exception.
        raise jexc.get_py4jdbc_exception() from py4j_exc


def get_jvm_exc_data(java_exc, gateway_client):
    exc_data = gateway_client.entry_point.getExceptionData(java_exc)
    exc_data = pickle.loads(exc_data)
    return _JavaException(java_exc, *exc_data)


class _JavaException:

    def __init__(self, java_exc, msg, traceback, classlist):
        self.java_exc = java_exc
        self.msg = msg
        self.tb = traceback
        getclassname = lambda s: s.split()[1]
        self.classlist = tuple(list(map(getclassname, classlist)))

    def get_py4jdbc_exception(self):
        exc = self.ExceptionClass(self)
        return exc

    def is_java_sql_error(self):
        return 'java.sql.SQLException' in self.classlist

    def tb_stripped(self):
        _, tb = self.tb.split(':', 1)
        return tb.strip()

    @CachedAttr
    def sqlstate(self):
        return self.java_exc.getSQLState()

    @property
    def sqlstate_class(self):
        return self.sqlstate[:2]

    @CachedAttr
    def errorcode(self):
        return self.java_exc.getErrorCode()

    @CachedAttr
    def javalib(self):
        '''I.e., returns org.postgresql from org.postgresql.util.PSQLError
        '''
        return re.match(r'\w+\.\w+', self.classlist[0]).group()

    @CachedAttr
    def exc_messages(self):
        sqlstate_mod = 'py4jdbc.exceptions.sqlstate'
        exc_module = _java_error_type_map.get(self.javalib, sqlstate_mod)
        return resolve_name('messages', exc_module)

    @CachedAttr
    def exc_classes(self):
        exc_module = _java_error_type_map[self.javalib]
        return resolve_name('exc_classes', exc_module)

    @property
    def ExceptionClass(self):
        '''Retrieve a custom exception class, if any, otherwise
        return DatabaseError.
        '''
        return self.exc_classes.get(self.sqlstate_class, DatabaseError)




import os
import sys
import signal
import atexit
import logging
import threading
from subprocess import Popen, PIPE
from os.path import abspath, dirname, join

from py4j.java_gateway import GatewayClient, JavaGateway

import py4jdbc
from py4jdbc.version import __version__ as py4jdbc_version


class GatewayProcess:
    JVM_MEMORY = os.environ.get('PY4JDBC_GATEWAY_MEMORY', '512m')
    command = ['java', '-Xmx%s' % JVM_MEMORY, 'Gateway']

    shutdown_signals = (
        signal.SIGINT, signal.SIGTERM,
        signal.SIGHUP, signal.SIGQUIT,
        signal.SIGABRT)

    def __init__(self, handle_signals=True, require_ctx_manager=True):
        self._gateway = None
        self._in_ctx_manager = False
        self._require_ctx_manager = require_ctx_manager
        self._shutdown_event = threading.Event()
        self._handle_signals = handle_signals
        self.logger = logging.getLogger('py4jdbc')
        self.is_running = False

    def __enter__(self):
        self._in_ctx_manager = True
        return self.gateway

    def __exit__(self, *args):
        self.shutdown()

    def run(self):
        if self._gateway is not None:
            msg = 'GatewayProcess is already running: %r'
            raise RuntimeError(msg % self)

        if self._require_ctx_manager and not self._in_ctx_manager:
            msg = 'The GatewayProcess must be invoked as a context manager.'
            raise RuntimeError(msg)

        # If program terminates normally, shut down the gateway.
        atexit.register(self.shutdown)

        # Establish signals so if it fails, shutdown the gateway.
        if self._handle_signals:
            self._set_signal_handlers()

        # Now launch the gateway server.
        try:
            self._launch_gateway()
        except Exception as exc:
            self.shutdown()
            raise exc
        self._lauch_echo_thread()
        client = GatewayClient(port=self._gateway_port)
        gateway = JavaGateway(client, auto_convert=True)

        self.is_running = True
        return gateway

    def shutdown(self):
        if self.is_running:
            self.logger.info('Shutting down gateway server: %r', self)
            # Tell the echo server to stop.
            self._shutdown_event.set()
            # Try shutting down the gateway server.
            self.gateway.shutdown()
            # Then forcibly kill.
            os.killpg(os.getpgid(self._proc.pid), signal.SIGTERM)
            self.is_running = False

    # -----------------------------------------------------------------------
    # Handle signals.
    # -----------------------------------------------------------------------
    def _shutdown_handler(self, signal, frame):
        self.shutdown()

    def _set_signal_handlers(self):
        self.logger.info('Setting shutdown signal handlers.')
        default_int_handler = signal.default_int_handler
        for sig in self.shutdown_signals:
            handler = signal.getsignal(sig)
            if handler is self._shutdown_handler:
                continue
            elif handler not in [0, default_int_handler]:
                def _handler(*args, **kwargs):
                    try:
                        handler(*args, **kwargs)
                    except RecursionError:
                        return
                    default_int_handler(*args, **kwargs)
                handler = _handler
            else:
                handler = self._shutdown_handler
            signal.signal(sig, handler)

    # -----------------------------------------------------------------------
    # Runs the gateway subprocess and streams it's output to stdout.
    # -----------------------------------------------------------------------
    @property
    def gateway(self):
        if self._gateway is None:
            self._gateway = self.run()
        return self._gateway

    def _launch_gateway(self):
        self.logger.info('Launching gateway server.')
        # Start the GatewayServer.
        self._proc = proc = Popen(
            self.command, stdout=PIPE, stdin=PIPE, preexec_fn=os.setsid)
        try:
            # Determine which ephemeral port the server started on:
            gateway_port = gateway_port = proc.stdout.readline()
            if isinstance(gateway_port, bytes):
                gateway_port = gateway_port.decode('ascii')
            self._gateway_port = gateway_port = int(gateway_port)
            self.logger.info('Gateway server port is %s', gateway_port)
        except ValueError:
            # Grab the remaining lines of stdout
            (stdout, _) = proc.communicate()
            if isinstance(stdout, bytes):
                stdout = stdout.decode('utf8')
            exit_code = proc.poll()
            error_msg = "Launching GatewayServer failed"
            error_msg += " with exit code %d!\n" % exit_code if exit_code else "!\n"
            error_msg += "Warning: Expected GatewayServer to output a port, but found "
            if gateway_port == "" and stdout == "":
                error_msg += "no output.\n"
            else:
                error_msg += "the following:\n\n"
                error_msg += "--------------------------------------------------------------\n"
                error_msg += gateway_port + stdout
                error_msg += "--------------------------------------------------------------\n"
            raise Exception(error_msg)

    def _lauch_echo_thread(self):
        self.logger.info('Launching gateway server echo thread.')
        out = self._proc.stdout
        event = self._shutdown_event
        _EchoOutputThread(out, event).start()


class _EchoOutputThread(threading.Thread):

    def __init__(self, stream, shutdown_event):
        threading.Thread.__init__(self)
        self.daemon = True
        self.stream = stream
        self.shutdown_event = shutdown_event

    def run(self):
        while True:
            if self.shutdown_event.isSet():
                sys.exit(0)
            line = self.stream.readline()
            if isinstance(line, bytes):
                line = line.decode('utf8')
            sys.stderr.write(line)



from py4jdbc.dbapi2 import *
from py4jdbc.exceptions.dbapi2 import *


py4jdbc_connect = connect

# DB-API 2.0 Module Interface connect constructor
def connect(host, port=None, database=None, user=None, password=None, gateway=None):
    if port is not None:
        jdbc_url = 'jdbc:vertica://%s:%s/%s' % (host, port, database)
    else:
        jdbc_url = 'jdbc:vertica://%s/%s'
        jdbc_url = jdbc_url % (host, database)
    return py4jdbc_connect(jdbc_url, user, password, gateway)




from collections import ChainMap

from py4jdbc.exceptions import dbapi2
from py4jdbc.exceptions import sqlstate


exc_classes = ChainMap(sqlstate.exc_classes).new_child()


class _DerbyCodeAgg(dbapi2.CodeAggregatorMeta):
    exc_classes = exc_classes


class ExecutionError(dbapi2.DatabaseError, metaclass=_DerbyCodeAgg):
    code = 'F0'



from py4jdbc.exceptions import dbapi2


exc_classes = {}


class _SqlstateMeta(dbapi2.CodeAggregatorMeta):
    exc_classes = exc_classes


class Warning(dbapi2.Warning, metaclass=_SqlstateMeta):
    code = '01'

class NoData(dbapi2.Warning, metaclass=_SqlstateMeta):
    code = '02'

class SqlStatementNotYetComplete(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '03'

class ConnectionError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '08'

class TriggeredActionError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '09'

class FeatureNotSupported(dbapi2.NotSupportedError, metaclass=_SqlstateMeta):
    code = '0A'

class InvalidTransactionInitiation(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '0B'

class LocatorError(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '0F'

class InvalidGrantor(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '0L'

class InvalidRoleSpecification(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '0P'

class DiagnosticsError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '0Z'

class CaseNotFound(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '20'

class CardinalityViolation(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '21'

class DataError(dbapi2.DataError, metaclass=_SqlstateMeta):
    code = '22'

class IntegrityConstraintViolation(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '23'

class InvalidCursorState(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '24'

class InvalidTransactionState(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '25'

class InvalidSQLStatementName(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '26'

class TriggeredDataChangeViolation(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '27'

class InvalidAuthorizationSpecification(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '28'

class DependentPrivilegeDescriptorsStillExist(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '2B'

class InvalidTransactionTermination(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '2D'

class SqlRoutineError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '2F'

class InvalidCursorName(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '34'

class ExternalRoutineError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '38'

class ExternalRoutineInvocationError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '39'

class SavepointError(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '3B'

class InvalidCatalogName(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '3D'

class InvalidSchemaName(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '3F'

class TransactionRollback(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '40'

class SyntaxOrAccessRuleViolationError(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '42'

class WithCheckOptionViolation(dbapi2.ProgrammingError, metaclass=_SqlstateMeta):
    code = '44'

class InsufficientResources(dbapi2.OperationalError, metaclass=_SqlstateMeta):
    code = '53'

class ProgramLimitExceeded(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '54'

class ObjectNotInPrerequisiteState(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '55'

class OperatorIntervention(dbapi2.DatabaseError, metaclass=_SqlstateMeta):
    code = '57'

class SystemError(dbapi2.OperationalError, metaclass=_SqlstateMeta):
    code = '58'


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import subprocess

from setuptools import setup
from distutils.command.build import build
from distutils.command.clean import clean
from distutils.spawn import find_executable

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ["py4j==0.10.1"]
setup_requirements = ["pytest-runner==2.9", "wheel"]
test_requirements = ["pytest==2.9.2", "coverage==4.1", "pytest-cov==2.3.0"]

exec(compile(open("py4jdbc/version.py").read(), "py4jdbc/version.py", 'exec'))
VERSION = __version__  # noqa

class jar_build(build):
    def run(self):
        """
        Compile the companion jar file.
        """

        if find_executable('sbt') is None:
            raise EnvironmentError("""

The executable "sbt" cannot be found.

Please install the "sbt" tool to build the companion jar file.
""")

        build.run(self)

        cwd = os.getcwd()
        os.chdir('py4jdbc/scala')
        subprocess.check_call('sbt assembly', shell=True)
        os.chdir(cwd)

class jar_clean(clean):
    def run(self):
        """
        Cleans the scala targets from the system.
        """
        clean.run(self)

        cwd = os.getcwd()
        os.chdir('py4jdbc/scala')
        subprocess.check_call('sbt clean', shell=True)
        os.chdir(cwd)

setup(
    name='py4jdbc',
    version=VERSION,
    description="py4j JDBC wrapper",
    long_description=readme,
    author="Thom Neale",
    author_email='tneale@massmutual.com',
    url='https://github.com/massmutual/py4jdbc',
    packages=['py4jdbc', 'py4jdbc.exceptions'],
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords=['jdbc', 'dbapi', 'py4j'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    cmdclass={
        'build': jar_build,
        'clean': jar_clean
    },
    package_data={
        'py4jdbc': [
            'scala/build.sbt',
            'scala/LICENCE',
            'scala/project/assembly.sbt',
            'scala/src/main/scala/GatewayServer.scala'
        ]
    },
    data_files=[
        ('share/py4jdbc', ["py4jdbc/scala/target/scala-2.10/py4jdbc-assembly-{0}.jar".format(__version__)])
    ],
    setup_requires=setup_requirements
)



from collections import ChainMap

from py4jdbc.exceptions import dbapi2


exc_classes = ChainMap(sqlstate.exc_classes).new_child()


class _VCodeAgg(dbapi2.CodeAggregatorMeta):
    exc_classes = exc_classes


class VerticaMultinodeError(dbapi2.DatabaseError, metaclass=_VCodeAgg):
    code = 'V1'

class VerticaMiscellaneousError(dbapi2.DatabaseError, metaclass=_VCodeAgg):
    code = 'V2'

class ConfigurationFileError(dbapi2.DatabaseError, metaclass=_VCodeAgg):
    code = 'VC'

class DbDesignerError(dbapi2.DatabaseError, metaclass=_VCodeAgg):
    code = 'VD'

class UserProcedureError(dbapi2.DatabaseError, metaclass=_VCodeAgg):
    code = 'VP'

class InternalError(dbapi2.DatabaseError, metaclass=_VCodeAgg):
    code = 'VX'


messages = (
    ('00000', 'ERRCODE_SUCCESSFUL_COMPLETION'),
    ('01000', 'ERRCODE_WARNING'),
    ('01003', 'ERRCODE_WARNING_NULL_VALUE_ELIMINATED_IN_SET_FUNCTION'),
    ('01004', 'ERRCODE_WARNING_STRING_DATA_RIGHT_TRUNCATION'),
    ('01006', 'ERRCODE_WARNING_PRIVILEGE_NOT_REVOKED'),
    ('01007', 'ERRCODE_WARNING_PRIVILEGE_NOT_GRANTED'),
    ('01008', 'ERRCODE_WARNING_PRIVILEGE_ALREADY_GRANTED'),
    ('01009', 'ERRCODE_WARNING_PRIVILEGE_ALREADY_REVOKED'),
    ('0100C', 'ERRCODE_WARNING_DYNAMIC_RESULT_SETS_RETURNED'),
    ('01V01', 'ERRCODE_WARNING_DEPRECATED_FEATURE'),
    ('01V02', 'ERRCODE_WARNING_QUERY_RETRIED'),
    ('02000', 'ERRCODE_NO_DATA'),
    ('02001', 'ERRCODE_NO_ADDITIONAL_DYNAMIC_RESULT_SETS_RETURNED'),
    ('03000', 'ERRCODE_SQL_STATEMENT_NOT_YET_COMPLETE'),
    ('08000', 'ERRCODE_CONNECTION_EXCEPTION'),
    ('08001', 'ERRCODE_SQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION'),
    ('08003', 'ERRCODE_CONNECTION_DOES_NOT_EXIST'),
    ('08004', 'ERRCODE_SQLSERVER_REJECTED_ESTABLISHMENT_OF_SQLCONNECTION'),
    ('08006', 'ERRCODE_CONNECTION_FAILURE'),
    ('08007', 'ERRCODE_TRANSACTION_RESOLUTION_UNKNOWN'),
    ('08V01', 'ERRCODE_PROTOCOL_VIOLATION'),
    ('09000', 'ERRCODE_TRIGGERED_ACTION_EXCEPTION'),
    ('0A000', 'ERRCODE_FEATURE_NOT_SUPPORTED'),
    ('0A005', 'ERRCODE_PLAN_TO_SQL_NOT_SUPPORTED'),
    ('0B000', 'ERRCODE_INVALID_TRANSACTION_INITIATION'),
    ('0F000', 'ERRCODE_LOCATOR_EXCEPTION'),
    ('0F001', 'ERRCODE_L_E_INVALID_SPECIFICATION'),
    ('0L000', 'ERRCODE_INVALID_GRANTOR'),
    ('0LV01', 'ERRCODE_INVALID_GRANT_OPERATION'),
    ('0P000', 'ERRCODE_INVALID_ROLE_SPECIFICATION'),
    ('21000', 'ERRCODE_CARDINALITY_VIOLATION'),
    ('22000', 'ERRCODE_DATA_EXCEPTION'),
    ('22001', 'ERRCODE_STRING_DATA_RIGHT_TRUNCATION'),
    ('22002', 'ERRCODE_NULL_VALUE_NO_INDICATOR_PARAMETER'),
    ('22003', 'ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE'),
    ('22004', 'ERRCODE_NULL_VALUE_NOT_ALLOWED'),
    ('22005', 'ERRCODE_ERROR_IN_ASSIGNMENT'),
    ('22007', 'ERRCODE_INVALID_DATETIME_FORMAT'),
    ('22008', 'ERRCODE_DATETIME_FIELD_OVERFLOW'),
    ('22009', 'ERRCODE_INVALID_TIME_ZONE_DISPLACEMENT_VALUE'),
    ('2200B', 'ERRCODE_ESCAPE_CHARACTER_CONFLICT'),
    ('2200C', 'ERRCODE_INVALID_USE_OF_ESCAPE_CHARACTER'),
    ('2200D', 'ERRCODE_INVALID_ESCAPE_OCTET'),
    ('2200F', 'ERRCODE_ZERO_LENGTH_CHARACTER_STRING'),
    ('2200G', 'ERRCODE_MOST_SPECIFIC_TYPE_MISMATCH'),
    ('22010', 'ERRCODE_INVALID_INDICATOR_PARAMETER_VALUE'),
    ('22011', 'ERRCODE_SUBSTRING_ERROR'),
    ('22012', 'ERRCODE_DIVISION_BY_ZERO'),
    ('22015', 'ERRCODE_INTERVAL_FIELD_OVERFLOW'),
    ('22018', 'ERRCODE_INVALID_CHARACTER_VALUE_FOR_CAST'),
    ('22019', 'ERRCODE_INVALID_ESCAPE_CHARACTER'),
    ('2201B', 'ERRCODE_INVALID_REGULAR_EXPRESSION'),
    ('2201E', 'ERRCODE_INVALID_ARGUMENT_FOR_LOG'),
    ('2201F', 'ERRCODE_INVALID_ARGUMENT_FOR_POWER_FUNCTION'),
    ('2201G', 'ERRCODE_INVALID_ARGUMENT_FOR_WIDTH_BUCKET_FUNCTION'),
    ('22020', 'ERRCODE_INVALID_LIMIT_VALUE'),
    ('22021', 'ERRCODE_CHARACTER_NOT_IN_REPERTOIRE'),
    ('22022', 'ERRCODE_INDICATOR_OVERFLOW'),
    ('22023', 'ERRCODE_INVALID_PARAMETER_VALUE'),
    ('22024', 'ERRCODE_UNTERMINATED_C_STRING'),
    ('22025', 'ERRCODE_INVALID_ESCAPE_SEQUENCE'),
    ('22026', 'ERRCODE_STRING_DATA_LENGTH_MISMATCH'),
    ('22027', 'ERRCODE_TRIM_ERROR'),
    ('2202E', 'ERRCODE_ARRAY_ELEMENT_ERROR'),
    ('22906', 'ERRCODE_NONSTANDARD_USE_OF_ESCAPE_CHARACTER'),
    ('22V01', 'ERRCODE_FLOATING_POINT_EXCEPTION'),
    ('22V02', 'ERRCODE_INVALID_TEXT_REPRESENTATION'),
    ('22V03', 'ERRCODE_INVALID_BINARY_REPRESENTATION'),
    ('22V04', 'ERRCODE_BAD_COPY_FILE_FORMAT'),
    ('22V05', 'ERRCODE_UNTRANSLATABLE_CHARACTER'),
    ('22V0B', 'ERRCODE_ESCAPE_CHARACTER_ON_NOESCAPE'),
    ('22V21', 'ERRCODE_INVALID_EPOCH'),
    ('22V22', 'ERRCODE_PLPGSQL_ERROR'),
    ('22V23', 'ERRCODE_RAISE_EXCEPTION'),
    ('22V24', 'ERRCODE_COPY_PARSE_ERROR'),
    ('23000', 'ERRCODE_INTEGRITY_CONSTRAINT_VIOLATION'),
    ('23001', 'ERRCODE_RESTRICT_VIOLATION'),
    ('23502', 'ERRCODE_NOT_NULL_VIOLATION'),
    ('23503', 'ERRCODE_FOREIGN_KEY_VIOLATION'),
    ('23505', 'ERRCODE_UNIQUE_VIOLATION'),
    ('23514', 'ERRCODE_CHECK_VIOLATION'),
    ('24000', 'ERRCODE_INVALID_CURSOR_STATE'),
    ('25000', 'ERRCODE_INVALID_TRANSACTION_STATE'),
    ('25001', 'ERRCODE_ACTIVE_SQL_TRANSACTION'),
    ('25002', 'ERRCODE_BRANCH_TRANSACTION_ALREADY_ACTIVE'),
    ('25003', 'ERRCODE_INAPPROPRIATE_ACCESS_MODE_FOR_BRANCH_TRANSACTION'),
    ('25004', 'ERRCODE_INAPPROPRIATE_ISOLATION_LEVEL_FOR_BRANCH_TRANSACTION'),
    ('25005', 'ERRCODE_NO_ACTIVE_SQL_TRANSACTION_FOR_BRANCH_TRANSACTION'),
    ('25006', 'ERRCODE_READ_ONLY_SQL_TRANSACTION'),
    ('25007', 'ERRCODE_SCHEMA_AND_DATA_STATEMENT_MIXING_NOT_SUPPORTED'),
    ('25008', 'ERRCODE_HELD_CURSOR_REQUIRES_SAME_ISOLATION_LEVEL'),
    ('25V01', 'ERRCODE_NO_ACTIVE_SQL_TRANSACTION'),
    ('25V02', 'ERRCODE_IN_FAILED_SQL_TRANSACTION'),
    ('26000', 'ERRCODE_INVALID_SQL_STATEMENT_NAME'),
    ('27000', 'ERRCODE_TRIGGERED_DATA_CHANGE_VIOLATION'),
    ('28000', 'ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION'),
    ('28001', 'ERRCODE_ACCOUNT_LOCKED'),
    ('28002', 'ERRCODE_PASSWORD_EXPIRED'),
    ('28003', 'ERRCODE_PASSWORD_IN_GRACE_PERIOD'),
    ('2B000', 'ERRCODE_DEPENDENT_PRIVILEGE_DESCRIPTORS_STILL_EXIST'),
    ('2BV01', 'ERRCODE_DEPENDENT_OBJECTS_STILL_EXIST'),
    ('2D000', 'ERRCODE_INVALID_TRANSACTION_TERMINATION'),
    ('2F000', 'ERRCODE_SQL_ROUTINE_EXCEPTION'),
    ('2F002', 'ERRCODE_S_R_E_MODIFYING_SQL_DATA_NOT_PERMITTED'),
    ('2F003', 'ERRCODE_S_R_E_PROHIBITED_SQL_STATEMENT_ATTEMPTED'),
    ('2F004', 'ERRCODE_S_R_E_READING_SQL_DATA_NOT_PERMITTED'),
    ('2F005', 'ERRCODE_S_R_E_FUNCTION_EXECUTED_NO_RETURN_STATEMENT'),
    ('34000', 'ERRCODE_INVALID_CURSOR_NAME'),
    ('38000', 'ERRCODE_EXTERNAL_ROUTINE_EXCEPTION'),
    ('38001', 'ERRCODE_E_R_E_CONTAINING_SQL_NOT_PERMITTED'),
    ('38002', 'ERRCODE_E_R_E_MODIFYING_SQL_DATA_NOT_PERMITTED'),
    ('38003', 'ERRCODE_E_R_E_PROHIBITED_SQL_STATEMENT_ATTEMPTED'),
    ('38004', 'ERRCODE_E_R_E_READING_SQL_DATA_NOT_PERMITTED'),
    ('39000', 'ERRCODE_EXTERNAL_ROUTINE_INVOCATION_EXCEPTION'),
    ('39001', 'ERRCODE_E_R_I_E_INVALID_SQLSTATE_RETURNED'),
    ('39004', 'ERRCODE_E_R_I_E_NULL_VALUE_NOT_ALLOWED'),
    ('39V01', 'ERRCODE_E_R_I_E_TRIGGER_PROTOCOL_VIOLATED'),
    ('39V02', 'ERRCODE_E_R_I_E_SRF_PROTOCOL_VIOLATED'),
    ('3B000', 'ERRCODE_SAVEPOINT_EXCEPTION'),
    ('3B001', 'ERRCODE_S_E_INVALID_SPECIFICATION'),
    ('3D000', 'ERRCODE_INVALID_CATALOG_NAME'),
    ('3F000', 'ERRCODE_INVALID_SCHEMA_NAME'),
    ('40000', 'ERRCODE_TRANSACTION_ROLLBACK'),
    ('40001', 'ERRCODE_T_R_SERIALIZATION_FAILURE'),
    ('40002', 'ERRCODE_T_R_INTEGRITY_CONSTRAINT_VIOLATION'),
    ('40003', 'ERRCODE_T_R_STATEMENT_COMPLETION_UNKNOWN'),
    ('40V01', 'ERRCODE_T_R_DEADLOCK_DETECTED'),
    ('42000', 'ERRCODE_SYNTAX_ERROR_OR_ACCESS_RULE_VIOLATION'),
    ('42501', 'ERRCODE_INSUFFICIENT_PRIVILEGE'),
    ('42601', 'ERRCODE_SYNTAX_ERROR'),
    ('42602', 'ERRCODE_INVALID_NAME'),
    ('42611', 'ERRCODE_INVALID_COLUMN_DEFINITION'),
    ('42622', 'ERRCODE_NAME_TOO_LONG'),
    ('42701', 'ERRCODE_DUPLICATE_COLUMN'),
    ('42702', 'ERRCODE_AMBIGUOUS_COLUMN'),
    ('42703', 'ERRCODE_UNDEFINED_COLUMN'),
    ('42704', 'ERRCODE_UNDEFINED_OBJECT'),
    ('42710', 'ERRCODE_DUPLICATE_OBJECT'),
    ('42712', 'ERRCODE_DUPLICATE_ALIAS'),
    ('42723', 'ERRCODE_DUPLICATE_FUNCTION'),
    ('42725', 'ERRCODE_AMBIGUOUS_FUNCTION'),
    ('42803', 'ERRCODE_GROUPING_ERROR'),
    ('42804', 'ERRCODE_DATATYPE_MISMATCH'),
    ('42809', 'ERRCODE_WRONG_OBJECT_TYPE'),
    ('42830', 'ERRCODE_INVALID_FOREIGN_KEY'),
    ('42846', 'ERRCODE_CANNOT_COERCE'),
    ('42883', 'ERRCODE_UNDEFINED_FUNCTION'),
    ('42939', 'ERRCODE_RESERVED_NAME'),
    ('42P20', 'ERRCODE_WINDOWING_ERROR'),
    ('42V01', 'ERRCODE_UNDEFINED_TABLE'),
    ('42V02', 'ERRCODE_UNDEFINED_PARAMETER'),
    ('42V03', 'ERRCODE_DUPLICATE_CURSOR'),
    ('42V04', 'ERRCODE_DUPLICATE_DATABASE'),
    ('42V05', 'ERRCODE_DUPLICATE_PSTATEMENT'),
    ('42V06', 'ERRCODE_DUPLICATE_SCHEMA'),
    ('42V07', 'ERRCODE_DUPLICATE_TABLE'),
    ('42V08', 'ERRCODE_AMBIGUOUS_PARAMETER'),
    ('42V09', 'ERRCODE_AMBIGUOUS_ALIAS'),
    ('42V10', 'ERRCODE_INVALID_COLUMN_REFERENCE'),
    ('42V11', 'ERRCODE_INVALID_CURSOR_DEFINITION'),
    ('42V12', 'ERRCODE_INVALID_DATABASE_DEFINITION'),
    ('42V13', 'ERRCODE_INVALID_FUNCTION_DEFINITION'),
    ('42V14', 'ERRCODE_INVALID_PSTATEMENT_DEFINITION'),
    ('42V15', 'ERRCODE_INVALID_SCHEMA_DEFINITION'),
    ('42V16', 'ERRCODE_INVALID_TABLE_DEFINITION'),
    ('42V17', 'ERRCODE_INVALID_OBJECT_DEFINITION'),
    ('42V18', 'ERRCODE_INDETERMINATE_DATATYPE'),
    ('42V21', 'ERRCODE_UNDEFINED_PROJECTION'),
    ('42V22', 'ERRCODE_UNDEFINED_NODE'),
    ('42V23', 'ERRCODE_UNDEFINED_PERMUTATION'),
    ('42V24', 'ERRCODE_UNDEFINED_USER'),
    ('42V25', 'ERRCODE_PATTERN_MATCH_ERROR'),
    ('42V26', 'ERRCODE_DUPLICATE_NODE'),
    ('44000', 'ERRCODE_WITH_CHECK_OPTION_VIOLATION'),
    ('53000', 'ERRCODE_INSUFFICIENT_RESOURCES'),
    ('53100', 'ERRCODE_DISK_FULL'),
    ('53200', 'ERRCODE_OUT_OF_MEMORY'),
    ('53300', 'ERRCODE_TOO_MANY_CONNECTIONS'),
    ('54000', 'ERRCODE_PROGRAM_LIMIT_EXCEEDED'),
    ('54001', 'ERRCODE_STATEMENT_TOO_COMPLEX'),
    ('54011', 'ERRCODE_TOO_MANY_COLUMNS'),
    ('54023', 'ERRCODE_TOO_MANY_ARGUMENTS'),
    ('55000', 'ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE'),
    ('55006', 'ERRCODE_OBJECT_IN_USE'),
    ('55V02', 'ERRCODE_CANT_CHANGE_RUNTIME_PARAM'),
    ('55V03', 'ERRCODE_LOCK_NOT_AVAILABLE'),
    ('55V04', 'ERRCODE_TM_MARKER_NOT_AVAILABLE'),
    ('57000', 'ERRCODE_OPERATOR_INTERVENTION'),
    ('57014', 'ERRCODE_QUERY_CANCELED'),
    ('57015', 'ERRCODE_SLOW_DELETE'),
    ('57V01', 'ERRCODE_ADMIN_SHUTDOWN'),
    ('57V02', 'ERRCODE_CRASH_SHUTDOWN'),
    ('57V03', 'ERRCODE_CANNOT_CONNECT_NOW'),
    ('58030', 'ERRCODE_IO_ERROR'),
    ('58V01', 'ERRCODE_UNDEFINED_FILE'),
    ('58V02', 'ERRCODE_DUPLICATE_FILE'),
    ('V1001', 'ERRCODE_LOST_CONNECTIVITY'),
    ('V1002', 'ERRCODE_K_SAFETY_VIOLATION'),
    ('V1003', 'ERRCODE_CLUSTER_CHANGE'),
    ('V2000', 'ERRCODE_AUTH_FAILED'),
    ('V2001', 'ERRCODE_LICENSE_ISSUE'),
    ('V2002', 'ERRCODE_MOVEOUT_ABORTED'),
    ('VC001', 'ERRCODE_CONFIG_FILE_ERROR'),
    ('VC002', 'ERRCODE_LOCK_FILE_EXISTS'),
    ('VD001', 'ERRCODE_DESIGNER_FUNCTION_ERROR'),
    ('VP000', 'ERRCODE_USER_PROC_ERROR'),
    ('VP001', 'ERRCODE_USER_PROC_EXEC_ERROR'),
    ('VX001', 'ERRCODE_INTERNAL_ERROR'),
    ('VX002', 'ERRCODE_DATA_CORRUPTED'),
    ('VX003', 'ERRCODE_INDEX_CORRUPTED'),
    ('VX004', 'ERRCODE_PLAN_TO_SQL_INTERNAL_EROR'),
)